
from angr import Analysis, register_analysis
from angr.analyses.reaching_definitions import OP_AFTER
from angr.analyses.reaching_definitions.external_codeloc import ExternalCodeLocation

from ..block import Block
from ..statement import Assignment
from ..expression import Tmp, Register


class Simplifier(Analysis):
    def __init__(self, block):
        """

        :param Block block:
        """

        self.block = block

        self.result_block = None

        self._analyze()

    def _analyze(self):

        block = self.block
        block = self._simplify_block_once(block)
        #print str(block)
        block = self._simplify_block_once(block)
        #print str(block)
        #block = self._simplify_block_once(block)
        #print str(block)

        self.result_block = block

    def _simplify_block_once(self, block):

        # reaching definition analysis
        rd = self.project.analyses.ReachingDefinitions(block=block)

        # propagator
        propagator = self.project.analyses.AILPropagator(block=block, reaching_definitions=rd)
        replacements = propagator._states.values()[0]._final_replacements
        new_block = self._replace_and_build(block, replacements)
        new_block = self._eliminate_dead_assignments(new_block)

        return new_block

    def _replace_and_build(self, block, replacements):

        new_statements = block.statements[::]

        for codeloc, old_expr, new_expr in replacements:
            stmt = new_statements[codeloc.stmt_idx]
            r, new_stmt = stmt.replace(old_expr, new_expr)

            if r:
                new_statements[codeloc.stmt_idx] = new_stmt

        new_block = block.copy()
        new_block.statements = new_statements

        return new_block

    def _eliminate_dead_assignments(self, block):

        new_statements = [ ]

        rd = self.project.analyses.ReachingDefinitions(block=block,
                                                       track_tmps=True,
                                                       observation_points=[ (block.statements[-1].ins_addr, OP_AFTER )]
                                                       )

        used_tmp_indices = set(rd.one_result.tmp_uses.keys())
        dead_virgins = rd.one_result._dead_virgin_definitions
        dead_virgins_stmt_idx = set([ d.codeloc.stmt_idx for d in dead_virgins
                                      if not isinstance(d.codeloc, ExternalCodeLocation) ])

        for idx, stmt in enumerate(block.statements):
            if type(stmt) is Assignment:
                if type(stmt.dst) is Tmp:
                    if stmt.dst.tmp_idx not in used_tmp_indices:
                        continue

                # is it a dead virgin?
                if idx in dead_virgins_stmt_idx:
                    continue

            new_statements.append(stmt)

        new_block = block.copy()
        new_block.statements = new_statements

        return new_block


register_analysis(Simplifier, 'AILSimplifier')




import logging

from angr import Analysis, register_analysis
from angr.analyses.reaching_definitions import OP_BEFORE
from angr.calling_conventions import SimRegArg, SimStackArg

from .. import Stmt, Expr

l = logging.getLogger('ailment.callsite_maker')


class CallSiteMaker(Analysis):
    """
    Add calling convention, declaration, and args to a call site.
    """
    def __init__(self, block):
        self.block = block

        self._reaching_definitions = None

        self.result_block = None

        self._analyze()

    def _analyze(self):

        last_stmt = self.block.statements[-1]

        if not type(last_stmt) is Stmt.Call:
            self.result_block = self.block
            return

        target = self._get_call_target(last_stmt)

        if target is None:
            return

        if target not in self.kb.functions:
            return

        func = self.kb.functions[target]

        if func.prototype is None:
            func.find_declaration()

        if func.prototype is None:
            # cannot find a declaration to it
            return

        # Make arguments
        args = [ ]
        if func.calling_convention is None:
            l.warning('%s has an unknown calling convention.', repr(func))
        else:
            arg_locs = func.calling_convention.arg_locs()
            for arg_loc in arg_locs:
                if type(arg_loc) is SimRegArg:
                    size = arg_loc.size
                    offset = arg_loc._fix_offset(None, size, arch=self.project.arch)
                    args.append(Expr.Register(None, None, offset, size * 8, reg_name=arg_loc.reg_name))
                else:
                    raise NotImplementedError('Not implemented yet.')

        new_stmts = self.block.statements[::]

        new_stmts[-1] = Stmt.Call(last_stmt, last_stmt.target,
                                  calling_convention=func.calling_convention,
                                  prototype=func.prototype,
                                  args=args,
                                  **last_stmt.tags
                                  )

        new_block = self.block.copy()
        new_block.statements = new_stmts

        self.result_block = new_block

    def _get_call_target(self, stmt):
        """

        :param Stmt.Call stmt:
        :return:
        """

        if type(stmt.target) is Expr.Const:
            return stmt.target.value

        return None

register_analysis(CallSiteMaker, 'AILCallSiteMaker')





class TaggedObject(object):
    """
    A class that takes arbitrary tags.
    """

    def __init__(self, **kwargs):
        self.tags = { }
        if kwargs:
            self.initialize_tags(kwargs)

    def initialize_tags(self, tags):
        for k, v in tags.iteritems():
            self.tags[k] = v

    def __getattr__(self, item):
        try:
            return self.tags[item]
        except KeyError:
            return super(TaggedObject, self).__getattribute__(item)
